@@ -29,6 +29,7 @@
 #include "vcs_version.h"
 #include "cli_config.h"
 
+#include <assert.h>
 #include <errno.h>
 #include <inttypes.h>
 #include <math.h>
@@ -139,6 +140,47 @@ static void print_stats(const int istty, const unsigned n, const unsigned num,
     fputs(buf, stderr);
 }
 
+static int picture_alloc(Dav1dPicture *const p, void *const _) {
+    const int hbd = p->p.bpc > 8;
+    const int aligned_w = (p->p.w + 127) & ~127;
+    const int aligned_h = (p->p.h + 127) & ~127;
+    const int has_chroma = p->p.layout != DAV1D_PIXEL_LAYOUT_I400;
+    const int ss_ver = p->p.layout == DAV1D_PIXEL_LAYOUT_I420;
+    const int ss_hor = p->p.layout != DAV1D_PIXEL_LAYOUT_I444;
+    ptrdiff_t y_stride = aligned_w << hbd;
+    ptrdiff_t uv_stride = has_chroma ? y_stride >> ss_hor : 0;
+    /* Due to how mapping of addresses to sets works in most L1 and L2 cache
+     * implementations, strides of multiples of certain power-of-two numbers
+     * may cause multiple rows of the same superblock to map to the same set,
+     * causing evictions of previous rows resulting in a reduction in cache
+     * hit rate. Avoid that by slightly padding the stride when necessary. */
+    if (!(y_stride & 1023))
+        y_stride += DAV1D_PICTURE_ALIGNMENT;
+    if (!(uv_stride & 1023) && has_chroma)
+        uv_stride += DAV1D_PICTURE_ALIGNMENT;
+    p->stride[0] = -y_stride;
+    p->stride[1] = -uv_stride;
+    const size_t y_sz = y_stride * aligned_h;
+    const size_t uv_sz = uv_stride * (aligned_h >> ss_ver);
+    const size_t pic_size = y_sz + 2 * uv_sz;
+
+    uint8_t *const buf = malloc(pic_size + DAV1D_PICTURE_ALIGNMENT * 2);
+    if (!buf) return DAV1D_ERR(ENOMEM);
+    p->allocator_data = buf;
+
+    const ptrdiff_t align_m1 = DAV1D_PICTURE_ALIGNMENT - 1;
+    uint8_t *const data = (uint8_t *)(((ptrdiff_t)buf + align_m1) & ~align_m1);
+    p->data[0] = data + y_sz - y_stride;
+    p->data[1] = has_chroma ? data + y_sz + uv_sz * 1 - uv_stride : NULL;
+    p->data[2] = has_chroma ? data + y_sz + uv_sz * 2 - uv_stride : NULL;
+
+    return 0;
+}
+
+static void picture_release(Dav1dPicture *const p, void *const _) {
+    free(p->allocator_data);
+}
+
 int main(const int argc, char *const *const argv) {
     const int istty = isatty(fileno(stderr));
     int res = 0;
@@ -162,6 +204,10 @@ int main(const int argc, char *const *const argv) {
     }
 
     parse(argc, argv, &cli_settings, &lib_settings);
+    if (cli_settings.neg_stride) {
+        lib_settings.allocator.alloc_picture_callback = picture_alloc;
+        lib_settings.allocator.release_picture_callback = picture_release;
+    }
 
     if ((res = input_open(&in, cli_settings.demuxer,
                           cli_settings.inputfile,
@@ -197,7 +243,6 @@ int main(const int argc, char *const *const argv) {
                     seq_skip);
     }
 
-    //getc(stdin);
     if (cli_settings.limit != 0 && cli_settings.limit < total)
         total = cli_settings.limit;
 
@@ -225,9 +270,10 @@ int main(const int argc, char *const *const argv) {
         memset(&p, 0, sizeof(p));
         if ((res = dav1d_send_data(c, &data)) < 0) {
             if (res != DAV1D_ERR(EAGAIN)) {
+                dav1d_data_unref(&data);
                 fprintf(stderr, "Error decoding frame: %s\n",
                         strerror(DAV1D_ERR(res)));
-                break;
+                if (res != DAV1D_ERR(EINVAL)) break;
             }
         }
 
@@ -235,7 +281,7 @@ int main(const int argc, char *const *const argv) {
             if (res != DAV1D_ERR(EAGAIN)) {
                 fprintf(stderr, "Error decoding frame: %s\n",
                         strerror(DAV1D_ERR(res)));
-                break;
+                if (res != DAV1D_ERR(EINVAL)) break;
             }
             res = 0;
         } else {
@@ -271,6 +317,7 @@ int main(const int argc, char *const *const argv) {
             if (res != DAV1D_ERR(EAGAIN)) {
                 fprintf(stderr, "Error decoding frame: %s\n",
                         strerror(DAV1D_ERR(res)));
+                if (res != DAV1D_ERR(EINVAL)) break;
             } else {
                 res = 0;
                 break;
